--  wwffwfwf - generates truth tables from logic expressions
--  Copyright (C) 2024 VegOwOtenks
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU Affero General Public License as published by
--  the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU Affero General Public License for more details.
--
--  You should have received a copy of the GNU Affero General Public License
--  along with this program.  If not, see <https://www.gnu.org/licenses/>.
--
--  contact me via vegowotenks at jossco dot de
module Lib
    ( parseFullString, evaluate, collectVariableNames, Expression, ParseError
    ) where

import Text.Parsec
import Text.Parsec.Expr
import Text.Parsec.Token
import Text.Parsec.Language

import Data.Set (Set)
import qualified Data.Set as Set

import Data.Map (Map)
import qualified Data.Map as Map

import Data.Functor.Identity

data Expression = Constant Bool | Binary BinaryOperator Expression Expression | Unary UnaryOperator Expression | Variable String
    deriving Show
data BinaryOperator = LogicalAnd | LogicalOr
    deriving Show
data UnaryOperator  = LogicalNot
    deriving Show

def :: GenLanguageDef String u Identity
def = emptyDef{ commentStart = ""
              , commentEnd = ""
              , identStart = letter <|> char '_'
              , identLetter = alphaNum <|> char '_'
              , opStart = oneOf "&|~"
              , opLetter = oneOf "&|~"
              , reservedOpNames = ["&", "|", "~"]
              , reservedNames = ["true", "false"]
              }

TokenParser{ parens = m_parens
           , identifier = m_identifier
           , reservedOp = m_reservedOp
           , reserved   = m_reserved
           , whiteSpace = m_whiteSpace } = makeTokenParser def

exprparser :: ParsecT String () Identity Expression
exprparser = buildExpressionParser table term <?> "expression"

table :: [[Operator String u Identity Expression]]
table = [
            [
                Prefix (m_reservedOp "~" >> return (Unary LogicalNot))
            ],
            [
                Infix (m_reservedOp "&" >> return (Binary LogicalAnd)) AssocLeft
            ],
            [
                Infix (m_reservedOp "|" >> return (Binary LogicalOr)) AssocLeft
            ]
        ]

term :: ParsecT String () Data.Functor.Identity.Identity Expression
term = m_parens exprparser
    <|> fmap Variable (m_identifier)
    <|> (m_reserved "true"  >> return (Constant True))
    <|> (m_reserved "false" >> return (Constant False))
    <?> "logical expression"

parseFullString :: SourceName -> String -> Either ParseError Expression
parseFullString = parse (m_whiteSpace *> exprparser <* eof)

collectVariableNames :: Expression -> Set String
collectVariableNames (Constant _)          = Set.empty
collectVariableNames (Binary _ left right) = Set.union (collectVariableNames left) (collectVariableNames right)
collectVariableNames (Unary _ right)       = collectVariableNames right
collectVariableNames (Variable name)       = Set.singleton name

evaluate :: Map String Bool -> Expression -> Bool
evaluate _  (Constant b)              = b
evaluate vs (Unary    LogicalNot e)   = not (evaluate vs e)
evaluate vs (Binary   LogicalAnd l r) = evaluate vs l && evaluate vs r
evaluate vs (Binary   LogicalOr  l r) = evaluate vs l || evaluate vs r
evaluate vs (Variable name)           = maybe False id (Map.lookup name vs)
